/* $Header: /home/felipe/rmarsa/source/rnpl/RCS/bbhutil.c,v 1.5 1996/11/14 19:42:02 rmarsa Exp $ */
/* HDF support routines */
/* parameter fetching routines */
/* These routines are written by Robert L. Marsa */
/* Copyright (c) 1995,1996 by Robert L. Marsa */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <mfhdf.h>
#include <stdarg.h>
#include <errno.h>
#include <bbhutil.h>
#include <sys/stat.h>

BBH_C *upcase(BBH_C *s)
{
   BBH_C *q;

   for(q=(BBH_C *)s;*q;q++)
      *q=toupper(*q);
   return s;
}

#ifndef HAVE_STRDUP
BBH_C *strdup(const BBH_C *s)
{
	BBH_C *t;
	
	if(s){
		t=(BBH_C *)malloc(sizeof(BBH_C)*(strlen(s)+1));
		if(t!=NULL)
			strcpy(t,s);
		return t;
	}else return NULL;
}
#endif

BBH_I strcmp_nc(const BBH_C *s1, const BBH_C *s2)
{
   BBH_C *q1,*q2;
   BBH_I r;

   q1=strdup(s1);
   q2=strdup(s2);
   upcase(q1);
   upcase(q2);
	 if(q1!=NULL && q2!=NULL)
		r=strcmp(q1,q2);
	else r=strcmp(s1,s2);
   if(q1) free(q1);
   if(q2) free(q2);
   return r;
}

BBH_V rdvcpy(BBH_D *r,BBH_D *s, BBH_I len)
{
	BBH_I i;

	if(r!=s && r && s)
		for(i=0;i<len;i++)
      r[i]=s[i];
}

/* returns a vector of size doubles;  prints error message if not enough memory */
BBH_D * vec_alloc(BBH_I size)
{
  BBH_D *q;

  q=(BBH_D *)malloc(sizeof(BBH_D)*size);
  if(q==NULL){
    fprintf(stderr,"Out of memory error in vec_alloc!\n");
		exit(0);
	}
  return q;
}

BBH_I * ivec_alloc(BBH_I size)
{
  BBH_I *q;

  q=(BBH_I *)malloc(sizeof(BBH_I)*size);
  if(q==NULL){
    fprintf(stderr,"Out of memory error in ivec_alloc!\n");
    exit(0);
  }
  return q;
}

BBH_V get_ivec_param(const BBH_C *p_file, BBH_C *name, BBH_I def,
                   BBH_I *p, BBH_I size)
{
  get_param_nc(p_file,name,def,"ivec",size,(BBH_V *)p);
}

BBH_V get_int_param(const BBH_C *p_file, BBH_C *name, BBH_I def,
                   BBH_I *p, BBH_I size)
{
  get_param_nc(p_file,name,def,"long",size,(BBH_V *)p);
}

BBH_V get_real_param(const BBH_C *p_file, BBH_C *name, BBH_I def,
                    BBH_D *p, BBH_I size)
{
  get_param_nc(p_file,name,def,"double",size,(BBH_V *)p);
}

BBH_V get_str_param(const BBH_C *p_file, BBH_C *name, BBH_I def,
                   BBH_C **p, BBH_I size)
{
  get_param_nc(p_file,name,def,"string",size,(BBH_V *)p);
}

/* search ascii file for a line of the form
   name := value.
   Assign value to p.
*/
#define BSIZE 4096
BBH_V get_param(const BBH_C *p_file, BBH_C *name, BBH_I def,
               BBH_C *type, BBH_I size, BBH_V *p)
{
  BBH_C string[BSIZE],temp[256],*r,*s;
  BBH_I found=0,i,v,iv,line=0;
  FILE *fp;
	BBH_I ltrace=0;
	
  fp=fopen(p_file,"r");
  if(fp==NULL){
    fprintf(stderr,"get_param: Unable to open file %s.\n",p_file);
    exit(0);
  }

  while(fgets(string,BSIZE,fp) && !found){
    line++;
    for(r=string;isspace(*r);r++);
    for(s=temp;!isspace(*r) && !(*r==':' && *(r+1)=='=');*(s++)=*(r++));
    *s='\0';
    found=!strcmp(temp,name);
    if(found){
      while(isspace(*r))
        r++;
      if(*r!=':' || *(r+1)!='='){
        found=0;
      }
    }
    if(found){
      r+=2;
      while(isspace(*r))
        r++;
      if(size>1 && strcmp("ivec",type)){
        if(*r!='['){
          fprintf(stderr,"get_param:Error:%d: %s is a vector.\n",line,name);
          exit(0);
        }
        r++;
      }
      if(!strcmp("double",type)){
        for(i=0;i<size;i++){
          while(isspace(*r))
            r++;
          v=sscanf(r,"%lf",(BBH_D *)p+i);
          if(v==0 || v==EOF){
            fprintf(stderr,"get_param:Error:%d:input too short, need %d floats.\n",line,size);
            exit(0);
          }
          while(*r && !isspace(*r))
            r++;
        }
      }else if(!strcmp("long",type)){
        for(i=0;i<size;i++){
          while(isspace(*r))
            r++;
          v=sscanf(r,"%d",(BBH_I *)p+i);
          if(v==0 || v==EOF){
            fprintf(stderr,"get_param:Error:%d:input too short, need %d ints.\n",line,size);
            exit(0);
          }
          while(*r && !isspace(*r))
            r++;
        }
      }else if(!strcmp("ivec",type)){
      	if(ltrace)
      		fprintf(stderr,"get_param: found ivec on line %d\n",line);
      	while(isspace(*r))
      		r++;
      	s=r;
      	iv=1;
state0:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state0\n");
      	if(*r=='*')
      		r++;
      	else if(isdigit(*r))
					while(isdigit(*(++r)));
				else goto error;
      	while(isspace(*r))
      	  r++;
state1:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state1\n");
      	if(*r!='-')
					goto error;
				r++;
      	while(isspace(*r))
      		r++;
state2:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state2\n");
      	if(*r=='*')
      		r++;
      	else if(isdigit(*r))
					while(isdigit(*(++r)));
				else goto error;
      	while(isspace(*r))
      	  r++;
state3:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state3\n");
      	if(*r=='/'){
      		r++;
      		while(isspace(*r))
      			r++;
      		goto state4;
      	}else if(*r==','){
      		r++;
      		iv++;
      		while(isspace(*r))
      			r++;
      		goto state0;
      	}else if(*r==NULL){
      		goto done;
      	}else goto error;
state4:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state4\n");
				if(!isdigit(*r))
					goto error;
				while(isdigit(*(++r)));
				while(isspace(*r))
					r++;
state5:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state5\n");
				if(*r==','){
					r++;
					iv++;
					while(isspace(*r))
						r++;
					goto state0;
				}else if(*r==NULL){
					goto done;
				}else goto error;
error:
      	fprintf(stderr,"get_param:Error:%d:ivec syntax error at %c.\n",line,*r);
      	exit(0);
done:   
      	if(ltrace){
      		fprintf(stderr,"get_param: entering done, iv=%d\n",iv);
      		for(i=0;i<size;i++)
      			fprintf(stderr,"get_param: p[%d]=%d\n",i,(*((BBH_I **)p))[i]);
      	}
				if(def && *((BBH_I **)p)!=NULL && size<3*iv+1){
					free(*((BBH_I **)p));
					*((BBH_I **)p)=ivec_alloc(3*iv+1);
				}
				(*((BBH_I **)p))[0]=iv;
				r=s;
				for(i=0;i<iv;i++){
					if(ltrace)
						fprintf(stderr,"get_param: entering loop with i=%d *r=%c\n",i,*r);
					if(*r=='*'){
						(*((BBH_I **)p))[3*i+1]=-1;
						r++;
					}else{
						sscanf(r,"%d",*((BBH_I **)p)+3*i+1);
          	while(isdigit(*(++r)));
          }
          if(ltrace)
          	fprintf(stderr,"get_param: got first num\n");
      		while(isspace(*r))
      	 		r++;
      	 	r++; /* skip - */
      		while(isspace(*r))
      	 		r++;
          if(ltrace)
          	fprintf(stderr,"get_param: skipped -\n");
					if(*r=='*'){
						(*((BBH_I **)p))[3*i+2]=-1;
						r++;
					}else{
						sscanf(r,"%d",*((BBH_I **)p)+3*i+2);
          	while(isdigit(*(++r)));
          }
          if(ltrace)
          	fprintf(stderr,"get_param: got second num\n");
      		while(isspace(*r))
      	 		r++;
      	 	if(*r=='/'){
      	 		r++;
      	 		while(isspace(*r))
      	 			r++;
      	 		sscanf(r,"%d",*((BBH_I **)p)+3*i+3);
      	 		while(isdigit(*(++r)));
          	if(ltrace)
          		fprintf(stderr,"get_param: got third num\n");
      	 		while(isspace(*r))
      	 			r++;
      	 	}
      	 	if(*r){
      	 		r++; /* skip , */
      	 		while(isspace(*r))
      	 			r++;
      	 	}
				}
      }else if(!strcmp("string",type)){
        for(i=0;i<size;i++){
          while(*r!='\"')
            r++;
          s=++r;
          while(*r && *r!='\"')
            r++;
          if(*r==NULL){
            fprintf(stderr,"get_param:Error:%d:input too short, need %d strings.\n",line,size);
            exit(0);
          }
          if(def && *((BBH_C **)p+i)!=NULL)
            free(*((BBH_C **)p+i));
          *((BBH_C **)p+i)=(BBH_C *)malloc(sizeof(BBH_C)*(r-s+1));
          strncpy(*((BBH_C **)p+i),s,r-s);
          (*((BBH_C **)p+i))[r-s]=0;
        }
      }
    }
  }
  if(!found)
    if(!def){
      fprintf(stderr,"get_param:Error: %s does not appear in parameter file.\n",name);
         fclose(fp);
      exit(0);
    }else fprintf(stderr,"get_param: Using default for %s\n",name);
  rewind(fp);
   fclose(fp);
}

/* same as above except not case sensitive */
BBH_V get_param_nc(const BBH_C *p_file, BBH_C *name, BBH_I def,
                     BBH_C *type, BBH_I size, BBH_V *p)
{
  BBH_C string[BSIZE],temp[256],*r,*s;
  BBH_I found=0,i,v,iv,line=0;
  FILE *fp;
	BBH_I ltrace=0;
	
  fp=fopen(p_file,"r");
  if(fp==NULL){
    fprintf(stderr,"get_param: Unable to open file %s.\n",p_file);
    exit(0);
  }

  while(fgets(string,BSIZE,fp) && !found){
    line++;
    for(r=string;isspace(*r);r++);
    for(s=temp;!isspace(*r) && !(*r==':' && *(r+1)=='=');*(s++)=*(r++));
    *s='\0';
    found=!strcmp_nc(temp,name);
    if(found){
      while(isspace(*r))
        r++;
      if(*r!=':' || *(r+1)!='='){
        found=0;
      }
      }
    if(found){
      r+=2;
      while(isspace(*r))
        r++;
      if(size>1 && strcmp("ivec",type)){
        if(*r!='['){
          fprintf(stderr,"get_param:Error:%d: %s is a vector.\n",line,name);
          exit(0);
        }
        r++;
      }
      if(!strcmp("double",type)){
        for(i=0;i<size;i++){
          while(isspace(*r))
            r++;
          v=sscanf(r,"%lf",(BBH_D *)p+i);
          if(v==0 || v==EOF){
            fprintf(stderr,"get_param:Error:%d:input too short, need %d floats.\n",line,size);
            exit(0);
          }
          while(*r && !isspace(*r))
            r++;
        }
      }else if(!strcmp("long",type)){
        for(i=0;i<size;i++){
          while(isspace(*r))
            r++;
          v=sscanf(r,"%d",(BBH_I *)p+i);
          if(v==0 || v==EOF){
            fprintf(stderr,"get_param:Error:%d:input too short, need %d ints.\n",line,size);
            exit(0);
          }
          while(*r && !isspace(*r))
            r++;
        }
      }else if(!strcmp("ivec",type)){
      	if(ltrace)
      		fprintf(stderr,"get_param: found ivec on line %d\n",line);
      	while(isspace(*r))
      		r++;
      	s=r;
      	iv=1;
state0:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state0\n");
      	if(*r=='*')
      		r++;
      	else if(isdigit(*r))
					while(isdigit(*(++r)));
				else goto error;
      	while(isspace(*r))
      	  r++;
state1:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state1\n");
      	if(*r!='-')
					goto error;
				r++;
      	while(isspace(*r))
      		r++;
state2:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state2\n");
      	if(*r=='*')
      		r++;
      	else if(isdigit(*r))
					while(isdigit(*(++r)));
				else goto error;
      	while(isspace(*r))
      	  r++;
state3:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state3\n");
      	if(*r=='/'){
      		r++;
      		while(isspace(*r))
      			r++;
      		goto state4;
      	}else if(*r==','){
      		r++;
      		iv++;
      		while(isspace(*r))
      			r++;
      		goto state0;
      	}else if(*r==NULL){
      		goto done;
      	}else goto error;
state4:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state4\n");
				if(!isdigit(*r))
					goto error;
				while(isdigit(*(++r)));
				while(isspace(*r))
					r++;
state5:
      	if(ltrace)
      		fprintf(stderr,"get_param: entering state5\n");
				if(*r==','){
					r++;
					iv++;
					while(isspace(*r))
						r++;
					goto state0;
				}else if(*r==NULL){
					goto done;
				}else goto error;
error:
      	fprintf(stderr,"get_param:Error:%d:ivec syntax error at %c.\n",line,*r);
      	exit(0);
done:   
      	if(ltrace){
      		fprintf(stderr,"get_param: entering done, iv=%d, size=%d\n",iv,size);
      		for(i=0;i<size;i++)
      			fprintf(stderr,"get_param: p[%d]=%d\n",i,((BBH_I *)p)[i]);
      	}
				if(size<3*iv+1){
					fprintf(stderr,"get_param:Error:%d:ivec too long for FORTRAN data type.\n",line);
					exit(0);
				}
				((BBH_I *)p)[0]=iv;
				r=s;
				for(i=0;i<iv;i++){
					if(ltrace)
						fprintf(stderr,"get_param: entering loop with i=%d *r=%c\n",i,*r);
					if(*r=='*'){
						((BBH_I *)p)[3*i+1]=-1;
						r++;
					}else{
						sscanf(r,"%d",((BBH_I *)p)+3*i+1);
          	while(isdigit(*(++r)));
          }
          if(ltrace)
          	fprintf(stderr,"get_param: got first num\n");
      		while(isspace(*r))
      	 		r++;
      	 	r++; /* skip - */
      		while(isspace(*r))
      	 		r++;
          if(ltrace)
          	fprintf(stderr,"get_param: skipped -\n");
					if(*r=='*'){
						((BBH_I *)p)[3*i+2]=-1;
						r++;
					}else{
						sscanf(r,"%d",((BBH_I *)p)+3*i+2);
          	while(isdigit(*(++r)));
          }
          if(ltrace)
          	fprintf(stderr,"get_param: got second num\n");
      		while(isspace(*r))
      	 		r++;
      	 	if(*r=='/'){
      	 		r++;
      	 		while(isspace(*r))
      	 			r++;
      	 		sscanf(r,"%d",((BBH_I *)p)+3*i+3);
      	 		while(isdigit(*(++r)));
          	if(ltrace)
          		fprintf(stderr,"get_param: got third num\n");
      	 		while(isspace(*r))
      	 			r++;
      	 	}
      	 	if(*r){
      	 		r++; /* skip , */
      	 		while(isspace(*r))
      	 			r++;
      	 	}
				}
      }else if(!strcmp("string",type)){
        for(i=0;i<size;i++){
          while(*r!='\"')
            r++;
          s=++r;
          while(*r && *r!='\"')
            r++;
          if(*r==NULL){
            fprintf(stderr,"get_param:Error:%d:input too short, need %d strings.\n",line,size);
            exit(0);
          }
          if(def && *((BBH_C **)p+i)!=NULL)
            free(*((BBH_C **)p+i));
          *((BBH_C **)p+i)=(BBH_C *)malloc(sizeof(BBH_C)*(r-s+1));
          strncpy(*((BBH_C **)p+i),s,r-s);
          (*((BBH_C **)p+i))[r-s]=0;
        }
      }
    }
  }
  if(!found)
    if(!def){
      fprintf(stderr,"get_param:Error: %s does not appear in parameter file.\n",name);
         fclose(fp);
      exit(0);
    }else fprintf(stderr,"get_param: Using default for %s\n",name);
  rewind(fp);
  fclose(fp);
}

static BBH_I call=0,len,top,mode;
static hdf_data *dfiles;
static BBH_I d_type_=DFNT_FLOAT64;

BBH_I gft_init(BBH_V)
{
	BBH_I ind;
	BBH_C *ev;
	int ltrace=0;
	
	if(!call){
		len=32;
		dfiles=NULL;
		if(!gft_newstruct(&dfiles,len)){
			fprintf(stderr,"gft_init: memory error\n");
			return(0);
		}else{
			call=1;
			top=ind=-1;
		}
		ev=getenv("GFT_HDFAPPEND");
		if(ev!=NULL)
			mode=0;
		else mode=1;
#ifdef HAVE_ATEXIT
		if(atexit(gft_close_all)!=0){
			fprintf(stderr,"gft_init: unable to register gft_close_all (atexit)\n");
		}
#endif
#ifdef HAVE_ATABORT
		if(atabort(gft_close_all)!=0){
			fprintf(stderr,"gft_init: unable to register gft_close_all (atabort)\n");
		}
#endif
	}
	return 1;
}

BBH_V gft_del(BBH_V)
{
	BBH_I i;
	
	if(call){
		for(i=0;i<top;i++){
			if(dfiles[i].state)
				SDend(dfiles[i].fid);
			if(dfiles[i].fn){
				free(dfiles[i].fn);
				dfiles[i].fn=NULL;
			}
			if(dfiles[i].ds){
				free(dfiles[i].ds);
				dfiles[i].ds=NULL;
			}
			if(dfiles[i].tmv){
				free(dfiles[i].tmv);
				dfiles[i].tmv=NULL;
			}
		}
		free(dfiles);
	}
}

BBH_V gft_clear(const BBH_I ind)
{
	BBH_I ltrace=0;

	if(ltrace != 0){
		fprintf(stderr,"gft_clear: call=%d  ind=%d top=%d\n",call,ind,top);
	}
	if(call && ind<=top){
		dfiles[ind].rank=0;
		if(dfiles[ind].ds){
			free(dfiles[ind].ds);
			dfiles[ind].ds=NULL;
		}
		if(dfiles[ind].tmv){
			free(dfiles[ind].tmv);
			dfiles[ind].tmv=NULL;
		}
	}
}

/* returns pointer to malloced string containing [a-zA-Z0-9_./] 
   or NULL upon failure */
BBH_C * gft_make_file_name(const BBH_C *s)
{
	BBH_C *n,*p;
	
	if(!s)
		n=NULL;
	else if((n=(BBH_C *)malloc(sizeof(BBH_C)*(strlen(s)+5)))==NULL){
		fprintf(stderr,"gft_make_file_name: can't get memory for new string.\n");
		n=NULL;
	}else{
		for(p=n;*s;s++)
			if(isalnum(*s) || *s=='_' || *s=='.' || *s=='/')
				*(p++)=*s;
		if(*(p-4)!='.' || *(p-3)!='h' || *(p-2)!='d' || *(p-1)!='f'){
			*(p++)='.';
			*(p++)='h';
			*(p++)='d';
			*(p++)='f';
		}
		*p=0;
	}
	return n;
}

/* search the array list for fn.  If found, return index, else return -1 */
BBH_I gft_search_list(const BBH_C *fn, hdf_data *list, BBH_I n)
{
	BBH_I i;
		
	if(list && fn && n<=top+1){
		for(i=0;i<n && strcmp(fn,list[i].fn);i++);
		if(i<n)
			return i;
		else return -1;
	}else return -1;
}

BBH_V gft_dump_hdf_data(hdf_data *df)
{
	BBH_I i;
	
	if(df){
		if(df->fn)
			fprintf(stderr,"fn=<%s>\n",df->fn);
		else fprintf(stderr,"fn=NULL\n");
		if(df->ds)
			fprintf(stderr,"ds=<%s>\n",df->ds);
		else fprintf(stderr,"ds=NULL\n");
		fprintf(stderr,"version=%d\n",df->vers);
		fprintf(stderr,"mode=%d\n",df->mode);
		fprintf(stderr,"fid=%d\n",df->fid);
		fprintf(stderr,"state=%d\n",df->state);
		fprintf(stderr,"sdsnum=%d\n",df->sdsnum);
		fprintf(stderr,"nslabs=%d\n",df->nslabs);
		fprintf(stderr,"rank=%d\n",df->rank);
	}
}

BBH_I gft_newstruct(hdf_data **df, BBH_I len)
{
	BBH_I i,ol;
	hdf_data *tmpd;
	
	if(*df){ /* already allocated */
		ol=len/2;
		tmpd=(hdf_data *)malloc(sizeof(hdf_data)*len);
		if(!tmpd){
			fprintf(stderr,"gft_newstruct: can't malloc new array.\n");
			errno=1;
			return(0);
		}
		for(i=0;i<ol;i++){
			tmpd[i].fn=(*df)[i].fn;
			tmpd[i].ds=(*df)[i].ds;
			tmpd[i].mode=(*df)[i].mode;
			tmpd[i].fid=(*df)[i].fid;
			tmpd[i].vers=(*df)[i].vers;
			tmpd[i].state=(*df)[i].state;
			tmpd[i].sdsnum=(*df)[i].sdsnum;
			tmpd[i].nslabs=(*df)[i].nslabs;
			tmpd[i].rank=(*df)[i].rank;
			tmpd[i].tmv=(*df)[i].tmv;
			tmpd[i].mtime=(*df)[i].mtime;
		}
		for(i=ol;i<len;i++){
			tmpd[i].fn=NULL;
			tmpd[i].ds=NULL;
			tmpd[i].mode=-1;
			tmpd[i].fid=-1;
			tmpd[i].vers=0;
			tmpd[i].state=0;
			tmpd[i].sdsnum=0;
			tmpd[i].nslabs=0;
			tmpd[i].rank=0;
			tmpd[i].tmv=NULL;
			tmpd[i].mtime=0;
		}
		free(*df);
		*df=tmpd;
		return(1);
	}else{ /* new */
		*df=(hdf_data *)malloc(sizeof(hdf_data)*len);
		if(!*df){
			fprintf(stderr,"gft_newstruct: can't malloc new array.\n");
			errno=1;
			return(0);
		}
		for(i=0;i<len;i++){
			(*df)[i].fn=NULL;
			(*df)[i].ds=NULL;
			(*df)[i].mode=-1;
			(*df)[i].fid=-1;
			(*df)[i].vers=0;
			(*df)[i].state=0;
			(*df)[i].sdsnum=0;
			(*df)[i].nslabs=0;
			(*df)[i].rank=0;
			(*df)[i].tmv=NULL;
			(*df)[i].mtime=0;
		}
		return(1);
	}
}

BBH_V gft_close_all(BBH_V)
{
	BBH_I i;
	
	if(call)
		for(i=0;i<=top;i++)
			if(dfiles[i].state){
				SDend(dfiles[i].fid);
				dfiles[i].state=0;
			}
}

BBH_I gft_set_d_type(BBH_I type)
{
	if(type==DFNT_FLOAT64 || type==DFNT_INT32){
		d_type_=type;
	}
	return d_type_;
}

BBH_I gft_check_file(BBH_I ind)
{
	BBH_I aind,v=1;
	int ltrace=0;
	
	if(ind<=top){
		if(SDfindattr(dfiles[ind].fid,"rnpl_file")==-1){
			fprintf(stderr,"gft_check_file: <%s> is not an RNPL file.\n",dfiles[ind].fn);
			errno=1;
			return 0;
		}
		if((aind=SDfindattr(dfiles[ind].fid,"rnpl_hdf_vers"))==-1){
			dfiles[ind].vers=1;
			if(ltrace)
				fprintf(stderr,"gft_check_file: version=1\n");
		}else{
			if(SDreadattr(dfiles[ind].fid,aind,&v)==-1){
				fprintf(stderr,"gft_check_file: can't read version from <%s>\n",dfiles[ind].fn);
				dfiles[ind].vers=1;
				if(ltrace)
					fprintf(stderr,"gft_check_file: version=1\n");
			}
			dfiles[ind].vers=v;
		}
		return 1;
	}
	return 0;
}

BBH_I gft_set_file_rnpl(BBH_I ind)
{
	BBH_I tmp=1;
	
	if(ind<=top){
		if(SDsetattr(dfiles[ind].fid,"rnpl_file",DFNT_INT32,1,(BBH_V *)&tmp)==-1){
			fprintf(stderr,"gft_set_file_rnpl: can't create attribute in <%s>\n",dfiles[ind].fn);
			return 0;
		}
		tmp=2;
		if(SDsetattr(dfiles[ind].fid,"rnpl_hdf_vers",DFNT_INT32,1,(BBH_V *)&tmp)==-1){
			fprintf(stderr,"gft_set_file_rnpl: can't create version in <%s>\n",dfiles[ind].fn);
			return 0;
		}
		return 1;
	}
	return 0;
}

/* create file and return index or -1 on failure */
/* if file has been accessed this run, don't overwrite */
/* if file exists and mode==0, don't overwrite */
BBH_I gft_create(const BBH_C *nm)
{
	BBH_I sr,amode,ind,nf=1;
	struct stat sbuf;
	int ltrace=0;
	
	if(ltrace)
		fprintf(stderr,"gft_create: call=%d nm=<%s>\n",call,nm);	
	
	if(!call){
		if(gft_init()){
			top=ind=0;
			dfiles[ind].fn=strdup(nm);
		}else{
			fprintf(stderr,"gft_create: memory error\n");
			return(-1);
		}
	}else{
		if((ind=gft_search_list(nm,dfiles,top+1))!=-1){
			dfiles[ind].sdsnum++;
			nf=0;
		}else{ /* new file */
			top++;
			if(top >= len){
			  len*=2;
				if(!gft_newstruct(&dfiles,len)){
					fprintf(stderr,"gft_create: memory error\n");
					return(-1);
				}
			}
			ind=top;
			dfiles[ind].fn=strdup(nm);
		}
	}
	if(!nf || !mode)
		amode=DFACC_RDWR;
	else amode=DFACC_CREATE;
	if(!nf)
		gft_clear(ind);
	if(!dfiles[ind].state){
		dfiles[ind].fid=SDstart(dfiles[ind].fn,amode);
		if(dfiles[ind].fid==-1 && amode==DFACC_CREATE){
			fprintf(stderr,"gft_create: unable to create file <%s>\n",
							dfiles[ind].fn);
			return(-1);
		}else if(dfiles[ind].fid==-1){
			amode=DFACC_CREATE;
			dfiles[ind].fid=SDstart(dfiles[ind].fn,amode);
			if(dfiles[ind].fid==-1){
				fprintf(stderr,"gft_create: unable to create file <%s>\n",
								dfiles[ind].fn);
				return(-1);
			}else{
				dfiles[ind].state=1;
				dfiles[ind].mode=amode;
			}
		}else{
			dfiles[ind].state=1;
			dfiles[ind].mode=amode;
		}
	}else if(dfiles[ind].mode!=DFACC_CREATE && dfiles[ind].mode!=DFACC_RDWR){ /* file already open */
		SDend(dfiles[ind].fid);
		dfiles[ind].state=0;
		dfiles[ind].fid=SDstart(dfiles[ind].fn,amode);
		if(dfiles[ind].fid==-1){
			fprintf(stderr,"gft_create: unable to create file <%s>\n",
							dfiles[ind].fn);
			return(-1);
		}else{
			dfiles[ind].state=1;
			dfiles[ind].mode=amode;
		}
	}
	if(!gft_set_file_rnpl(ind)){
		gft_close(nm);
		return(-1);
	}
	if((sr=stat(dfiles[ind].fn,&sbuf))==-1){
		fprintf(stderr,"gft_create: unable to get modification time for <%s>\n",
						dfiles[ind].fn);
		fprintf(stderr,"gft_create: stat returned %d\n",sr);
		return(-1);
	}else{
		dfiles[ind].mtime=sbuf.st_mtime;
	}
	return ind;
}

/* open file and return index or -1 on failure */
/* if file doesn't exist, create it */
BBH_I gft_open_rw(const BBH_C *nm)
{
	BBH_I ind,sr,nw=1;
	struct stat sbuf;
	int ltrace=0;
	
	if(ltrace)
		fprintf(stderr,"gft_open_rw: call=%d nm=<%s>\n",call,nm);	
	
	if(!call){
		if(gft_init()){
			top=ind=0;
			dfiles[ind].fn=strdup(nm);
		}else{
			fprintf(stderr,"gft_open_rw: memory error\n");
			return(-1);
		}
	}else{
		if((ind=gft_search_list(nm,dfiles,top+1))!=-1){
			dfiles[ind].sdsnum++;
		}else{ /* new file */
			top++;
			if(top >= len){
			  len*=2;
				if(!gft_newstruct(&dfiles,len)){
					fprintf(stderr,"gft_open_rw: memory error\n");
					return(-1);
				}
			}
			ind=top;
			dfiles[ind].fn=strdup(nm);
		}
	}
	if(!dfiles[ind].state){
		dfiles[ind].fid=SDstart(dfiles[ind].fn,DFACC_RDWR);
		if(dfiles[ind].fid==-1){
			fprintf(stderr,"gft_open_rw: unable to open file <%s>. Trying to create.\n",
							dfiles[ind].fn);
			return(gft_create(nm));
		}else{
			dfiles[ind].state=1;
			dfiles[ind].mode=DFACC_RDWR;
		}
		nw=0;
	}else if(dfiles[ind].mode!=DFACC_RDWR){ /* file already open */
		SDend(dfiles[ind].fid);
		dfiles[ind].state=0;
		dfiles[ind].fid=SDstart(dfiles[ind].fn,DFACC_RDWR);
		if(dfiles[ind].fid==-1){
			fprintf(stderr,"gft_open_rw: unable to open file <%s>\n",
							dfiles[ind].fn);
			return(-1);
		}else{
			dfiles[ind].state=1;
			dfiles[ind].mode=DFACC_RDWR;
		}
		nw=0;
	}
	if(!gft_check_file(ind)){
		gft_close(nm);
		return(-1);
	}
	if((sr=stat(dfiles[ind].fn,&sbuf))==-1){
		fprintf(stderr,"gft_open_rw: unable to get modification time for <%s>\n",
						dfiles[ind].fn);
		fprintf(stderr,"gft_open_rw: stat returned %d\n",sr);
		return(-1);
	}else{
		if(sbuf.st_mtime > dfiles[ind].mtime){
			gft_clear(ind);
			if(nw){
				SDend(dfiles[ind].fid);
				dfiles[ind].state=0;
				dfiles[ind].fid=SDstart(dfiles[ind].fn,DFACC_RDWR);
				if(dfiles[ind].fid==-1){
					fprintf(stderr,"gft_open_rw: unable to open file <%s>\n",
									dfiles[ind].fn);
					return(-1);
				}else{
					dfiles[ind].state=1;
					dfiles[ind].mode=DFACC_RDWR;
				}
			}
		}
		dfiles[ind].mtime=sbuf.st_mtime;
	}
	return ind;
}

/* open file and return index or -1 on failure */
BBH_I gft_open_rd(const BBH_C *nm)
{
	BBH_I ind,sr,nw=1;
	struct stat sbuf;
	int ltrace=0;
	
	if(ltrace)
		fprintf(stderr,"gft_open_rd: call=%d nm=<%s>\n",call,nm);	
	
	if(!call){
		if(gft_init()){
			top=ind=0;
			dfiles[ind].fn=strdup(nm);
		}else{
			fprintf(stderr,"gft_open_rd: memory error\n");
			return(-1);
		}
	}else{
		if((ind=gft_search_list(nm,dfiles,top+1))!=-1){
			dfiles[ind].sdsnum++;
		}else{ /* new file */
			top++;
			if(top >= len){
			  len*=2;
				if(!gft_newstruct(&dfiles,len)){
					fprintf(stderr,"gft_open_rd: memory error\n");
					return(-1);
				}
			}
			ind=top;
			dfiles[ind].fn=strdup(nm);
		}
	}
	if(!dfiles[ind].state){
		dfiles[ind].fid=SDstart(dfiles[ind].fn,DFACC_RDONLY);
		if(dfiles[ind].fid==-1){
			fprintf(stderr,"gft_open_rd: unable to open file <%s>\n",
							dfiles[ind].fn);
			return(-1);
		}else{
			dfiles[ind].state=1;
			dfiles[ind].mode=DFACC_RDONLY;
		}
		nw=0;
	}else if(dfiles[ind].mode!=DFACC_RDONLY){ /* file already open */
		SDend(dfiles[ind].fid);
		dfiles[ind].state=0;
		dfiles[ind].fid=SDstart(dfiles[ind].fn,DFACC_RDONLY);
		if(dfiles[ind].fid==-1){
			fprintf(stderr,"gft_open_rd: unable to open file <%s>\n",
							dfiles[ind].fn);
			return(-1);
		}else{
			dfiles[ind].state=1;
			dfiles[ind].mode=DFACC_RDONLY;
		}
		nw=0;
	}
	if(!gft_check_file(ind)){
		gft_close(nm);
		return(-1);
	}
	if((sr=stat(dfiles[ind].fn,&sbuf))==-1){
		fprintf(stderr,"gft_open_rd: unable to get modification time for <%s>\n",
						dfiles[ind].fn);
		fprintf(stderr,"gft_open_rd: stat returned %d\n",sr);
		return(-1);
	}else{
		if(ltrace != 0){
			fprintf(stderr,"gft_open_rd: sbuf.st_mtime=%d\n",sbuf.st_mtime);
			fprintf(stderr,"gft_open_rd: mtime=%d\n",dfiles[ind].mtime);
		}
		if(sbuf.st_mtime > dfiles[ind].mtime){
			gft_clear(ind);
			if(nw){
				SDend(dfiles[ind].fid);
				dfiles[ind].state=0;
				dfiles[ind].fid=SDstart(dfiles[ind].fn,DFACC_RDONLY);
				if(dfiles[ind].fid==-1){
					fprintf(stderr,"gft_open_rd: unable to open file <%s>\n",
									dfiles[ind].fn);
					return(-1);
				}else{
					dfiles[ind].state=1;
					dfiles[ind].mode=DFACC_RDONLY;
				}
			}
		}
		dfiles[ind].mtime=sbuf.st_mtime;
	}
	return ind;
}

BBH_I gft_close(const BBH_C *nm)
{
	BBH_I ind;
	const BBH_C *n;
	int ltrace=0;
	
	if(ltrace)
	  fprintf(stderr,"gft_close(%s)\n",nm);
	if(call){
		if((n=gft_make_file_name(nm))==NULL)
			n=nm;
		if((ind=gft_search_list(n,dfiles,top+1))!=-1){
			if(ltrace)
				fprintf(stderr,"gft_close: ind=%d\n",ind);
			if(dfiles[ind].state){
				if(ltrace)
					fprintf(stderr,"gft_close: calling SDend\n");
				SDend(dfiles[ind].fid);
				dfiles[ind].state=0;
			}
		}
		if(n!=nm)
			if(n) free((BBH_C *)n);
	}
}

/* read time level from hdf file.  Create storage */
BBH_I GFT_extract(const BBH_C *fname, /* in: file name */
					const BBH_I tlev, /* in: time level to retrieve (1 - ntlev) */
					BBH_C **func_name, /* name of this grid function */
					BBH_I *ntlev, /* number of time levels available in file */
					BBH_D **tvec, /* vector of time values */
					BBH_D *time, /* time of this slice */
					BBH_I *drank, /* rank of data set */
					BBH_I **shape, /* shape of data set */
					BBH_C ***cnames, /* names of coordinates */
					BBH_D ***coords, /* values of coordinates */
					BBH_D **data ) /* actual data */
{
	BBH_I fid, sds_id, nslabs;
	BBH_I start[MAXRANK],size[MAXRANK];
	BBH_I i,rank;
	BBH_I ind,sind,ln,sinc;
  BBH_I type,gattrs,attrs,sz,dim_id,dattrs;
  BBH_C name[60],*st;
	BBH_D t;
  int ltrace=0;
	
	if(tlev<0){ /* signal to free storage */
		if(*func_name)
			free(*func_name);
		if(*tvec)
			free(*tvec);
		if(*shape)
			free(*shape);
		if(*data)
			free(*data);
		for(i=0;i<*drank;i++){
			if(cnames[0][i])
				free(cnames[0][i]);
			if(coords[0][i])
				free(coords[0][i]);
		}
		if(*cnames)
			free(*cnames);
		if(*coords)
			free(*coords);
		return(1);
	}
	ind=gft_open_rd(fname);
  if(ind==-1){
    fprintf(stderr,"gft_extract: Unable to open <%s>.\n",fname);
    errno=1;
    return(0);
  }
	fid=dfiles[ind].fid;
	if( ltrace ) {
			fprintf(stderr,"gft_extract: fid: <%d>\n",fid);
			fprintf(stderr,"gft_extract: tmv: <%x>\n",dfiles[ind].tmv);
			fprintf(stderr,"gft_extract: vers: <%d>\n",dfiles[ind].vers);
	}
	if(dfiles[ind].tmv==NULL){ /* get file info */
		if(ltrace){
			fprintf(stderr,"gft_extract: initializing\n");
		}
		if(SDfileinfo(fid,(int32 *)&nslabs,(int32 *)&gattrs)==-1){
			fprintf(stderr,"gft_extract: Can't get file info.\n");
			errno=2;
			return(0);
		}
		dfiles[ind].nslabs=nslabs;
		if( ltrace ) {
			fprintf(stderr,"gft_extract: nslabs: <%d> gattrs: <%d>\n",nslabs,gattrs);
		}
		/* read first slice */
		sind=0;
	  if((sds_id=SDselect(fid,sind))==-1){
	    fprintf(stderr,"gft_extract: Can't select sds number %d.\n",sind);
	    errno=3;
	    return(0);
	  }
	  if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)size,(int32 *)&type,
								(int32 *)&attrs)==-1){
	    fprintf(stderr,"gft_extract: Can't get info about 1st sds\n");
	    errno=4;
	    return(0);
	  }
		dfiles[ind].rank=rank;
		if(type!=DFNT_FLOAT64){
			fprintf(stderr,"gft_extract: data set type is wrong.\n");
			errno=5;
			return(0);
		}
		if(attrs!=1){
			fprintf(stderr,"gft_extract: wrong type of HDF.\n");
			errno=6;
			return(0);
		}
		i=strlen(name);
		dfiles[ind].ds=(BBH_C *)malloc(sizeof(BBH_C)*(i+1));
		if(!dfiles[ind].ds){
			fprintf(stderr,"gft_extract: Can't allocate space for name.\n");
			errno=7;
			return(0);
		}
		strcpy(dfiles[ind].ds,name); /* copy and fix data set name */
		st=strrchr(dfiles[ind].ds,'[');
		if(st && *st=='[')
			*st='\0';
		if(dfiles[ind].vers==2){
			*ntlev=nslabs/(rank+1);
			sinc=rank+1;
		}else if(dfiles[ind].vers==1){
			*ntlev=nslabs-rank;
			sinc=1;
		}
		dfiles[ind].tmv=vec_alloc(*ntlev);
	  SDattrinfo(sds_id,0,name,(int32 *)&type,(int32 *)&sz);
		if(sz!=1 || type!=DFNT_FLOAT64){
			fprintf(stderr,"gft_extract: wrong type of HDF.\n");
			errno=8;
			return(0);
		}
		if(SDreadattr(sds_id,0,&t)==-1){
			fprintf(stderr,"gft_extract: Can't get time value.\n");
			errno=9;
			return(0);
		}
		dfiles[ind].tmv[0]=t;
		SDendaccess(sds_id);
		for(i=1,sind=rank+1;sind<nslabs;sind+=sinc,i++){
			if((sds_id=SDselect(fid,sind))==-1){
				fprintf(stderr,"gft_extract: Can't select sds number %d.\n",sind);
				errno=10;
				return(0);
			}
			SDreadattr(sds_id,0,&t);
			dfiles[ind].tmv[i]=t;
			SDendaccess(sds_id);
		}
	}
	rank=dfiles[ind].rank;
	nslabs=dfiles[ind].nslabs;
	if(dfiles[ind].vers==2){
		*ntlev=nslabs/(rank+1);
		sind=(tlev-1)*(rank+1);
	}else if(dfiles[ind].vers==1){
		*ntlev=nslabs-rank;
		sind=(tlev==1)?0:tlev+rank-1;
	}
	if(ltrace){
		fprintf(stderr,"gft_extract: rank=%d, nslabs=%d\n",rank,nslabs);
		fprintf(stderr,"gft_extract: version=%d\n",dfiles[ind].vers);
		fprintf(stderr,"gft_extract: *ntlev=%d, sind=%d\n",*ntlev,sind);
    gft_dump_hdf_data(&dfiles[ind]);
	}
	if(sind < 0 || sind >=nslabs){
		fprintf(stderr,"gft_extract: slice number %d out of range %d--%d\n",
            tlev,1,*ntlev);
		errno=11;
		return(0);
	}
	if((sds_id=SDselect(fid,sind))==-1){
		fprintf(stderr,"gft_extract: Can't select sds number %d.\n",sind);
		errno=12;
		return(0);
	}
	if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)size,(int32 *)&type,
							(int32 *)&attrs)==-1){
		fprintf(stderr,"gft_extract: Can't get info about sds %d\n",sind);
		errno=13;
		return(0);
	}
	for(ln=1,i=0;i<rank;i++)
		ln*=size[i];
	*func_name=(BBH_C *)malloc(sizeof(BBH_C)*(strlen(dfiles[ind].ds)+1));
	if(*func_name==NULL){
		fprintf(stderr,"gft_extract: Can't allocate space for data set name.\n");
		errno=14;
		return(0);
	}
	strcpy(*func_name,dfiles[ind].ds);
	*tvec=vec_alloc(*ntlev);
	rdvcpy(tvec[0],dfiles[ind].tmv,*ntlev);
	*drank=rank;
	*shape=ivec_alloc(rank);
	for(i=0;i<rank;i++){
		shape[0][i]=size[i];
		start[i]=0;
	}
	*cnames=(BBH_C **)malloc(sizeof(BBH_C *)*rank);
	if(*cnames==NULL){
		fprintf(stderr,"gft_extract: Can't allocate space for *cnames.\n");
		errno=16;
		return(0);
	}
	*coords=(BBH_D **)malloc(sizeof(BBH_D *)*rank);
	if(*coords==NULL){
		fprintf(stderr,"gft_extract: Can't allocate space for *coords.\n");
		errno=17;
		return(0);
	}	
	for(i=0;i<rank;i++){
		dim_id=SDgetdimid(sds_id,i);
		if(dim_id!=-1){
			if(SDdiminfo(dim_id,name,(int32 *)&sz,(int32 *)&type,(int32 *)&dattrs)==-1){
				fprintf(stderr,"gft_extract: Can't get dimension info.\n");
				errno=18;
				return(0);
			}
			if(type==0){
				fprintf(stderr,"gft_extract: Bad dimension %d.\n",i);
				errno=19;
				return(0);
			}
			cnames[0][i]=(BBH_C *)malloc(sizeof(BBH_C)*(strlen(name)+1));
			if(cnames[0][i]==NULL){
				fprintf(stderr,"gft_extract: Can't allocate memory for dim name %d.\n",i);
				errno=20;
				return(0);
			}
			st=strrchr(name,'[');
			if(st){
				strncpy(cnames[0][i],name,(BBH_I)(st-name));
				cnames[0][i][(BBH_I)(st-name)]=0;
			}else{
				strcpy(cnames[0][i],name);
			}
			if(sz!=size[i]){
				fprintf(stderr,"gft_extract: Dim %d has size %d while shape=%d.\n",i,sz,
								size[i]);
				errno=21;
				return(0);
			}
			coords[0][i]=(BBH_D *)malloc(sizeof(BBH_D)*sz);
			if(coords[0][i]==NULL){
				fprintf(stderr,"gft_extract: Can't allocate memory for dim %d.\n",i);
				errno=22;
				return(0);
			}
			if(SDgetdimscale(dim_id,coords[0][i])==-1){
				fprintf(stderr,"gft_extract: Can't read dim scale %d.\n",i);
				errno=23;
				return(0);
			}
		}else{
			fprintf(stderr,"gft_extract: Can't get dim id %d.\n",i);
			errno=24;
			return(0);
		}
	}
	if(SDreadattr(sds_id,0,&t)==-1){
		fprintf(stderr,"gft_extract: Can't get time value.\n");
		errno=25;
		return(0);
	}
	*time=t;
	*data=vec_alloc(ln);
	if(SDreaddata(sds_id,(int32 *)start,NULL,(int32 *)size,
								(BBH_V *)(*data))==-1){
		fprintf(stderr,"gft_extract: Can't read data.\n");
		errno=27;
		return(0);
	}
	SDendaccess(sds_id);
	return 1;
}

/* takes constant args and returns array */
BBH_D *gft_make_bbox(BBH_I rank, ...)
{
	va_list ap;
	BBH_I i,l;
	static BBH_D bb[2*MAXRANK];
	BBH_I ltrace=0;
	
	if(ltrace)
		fprintf(stderr,"gft_make_bbox: rank=%d\n",rank);
	if(rank>MAXRANK || rank<=0){
		fprintf(stderr,"gft_make_bbox: invalid rank=%d.\n",rank);
		return(NULL);
	}else{
		l=2*rank;
		va_start(ap,rank);
		for(i=0;i<l;i++){
			bb[i]=va_arg(ap,double);
			if(ltrace)
				fprintf(stderr,"gft_make_bbox: bb[%d]=%.16g\n",i,bb[i]);
		}
		va_end(ap);
		return(bb);
	}
}

/*	returns full function name:
			func_name(c0,c1,c2,...,t);
		rank is the spacial rank
		cnames[rank] contains the spacial coordinate names
*/
BBH_C *gft_make_full_func_name(BBH_C *func_name, BBH_I rank, BBH_C **cnames)
{
	BBH_C *n,*p;
	BBH_I len=0,j,i;
	
	if(!func_name || !cnames || rank<=0)
		n=NULL;
	else{ 
		len=strlen(func_name);
		for(i=0;i<rank;i++)
			len+=strlen(cnames[i]);
		if((n=(BBH_C *)malloc(sizeof(BBH_C)*(len+5)))==NULL){
			fprintf(stderr,
							"gft_make_full_func_name: can't get memory for new string.\n");
			n=NULL;
		}else{
			strcpy(n,func_name);
			i=strlen(n);
			n[i++]='(';
			for(j=0;j<rank;j++)
				i+=sprintf(n+i,"%s,",cnames[j]);
			sprintf(n+i,"t)");
		}
	}
	return n;	
}

BBH_I gft_write(const BBH_C *func_name, /* name of this grid function */
								BBH_D time, /* time of this slice */
								BBH_I *shape, /* shape of data set */
								BBH_I rank, /* rank of data set */
								BBH_D *data ) /* actual data */
{
	BBH_I ltrace=0;
	if(ltrace)
		fprintf(stderr,"gft_write: func_name=<%s> time=.16%g shape=%x rank=%d data=%x\n",
						func_name,time,shape,rank,data);
	return gft_write_brief(func_name,time,shape,rank,data);
}

BBH_I gft_write_brief(const BBH_C *func_name, /* name of this grid function */
											BBH_D time, /* time of this slice */
											BBH_I *shape, /* shape of data set */
											BBH_I rank, /* rank of data set */
											BBH_D *data ) /* actual data */
{
	BBH_C **cnames;
	BBH_D *coords,*c;
	BBH_I i,j,sz,ret;
	BBH_I ltrace=0;
	
	if(ltrace){
		fprintf(stderr,"gft_write_brief: rank=%d  shape=( ",rank);
		for(i=0;i<rank;i++)
			fprintf(stderr,"%d ",shape[i]);
		fprintf(stderr,")\n");
	}
	if(rank<=0){
		if(ltrace)
			fprintf(stderr,"gft_write_brief: calling write_full\n");
		return(gft_write_full(func_name,time,shape,cnames,rank,coords,data));
	}
	if((cnames=(BBH_C **)malloc(sizeof(BBH_C *)*(rank)))==NULL){
		fprintf(stderr,"gft_write_brief: can't get memory for coordinate names.\n");
		errno=9;
		return(0);
	}
	for(i=0;i<rank;i++){
		if((cnames[i]=(BBH_C *)malloc(sizeof(BBH_C)*2))==NULL){
			fprintf(stderr,"gft_write_brief: can't get memory for coordinate name.\n");
			errno=10;
			return(0);
		}
	}
	for(sz=0,i=0;i<rank;i++){
		cnames[i][0]='x'+i;
		cnames[i][1]=0;
		sz+=shape[i];
	}
	if(ltrace)
		fprintf(stderr,"gft_write_brief: trying to malloc %d doubles\n",sz);
	if((coords=(BBH_D *)malloc(sizeof(BBH_D)*sz))==NULL){
		fprintf(stderr,"gft_write_brief: can't get memory for coordinates.\n");
		errno=11;
		return(0);
	}
	for(c=coords,i=0;i<rank;i++){
		for(j=0;j<shape[i];j++,c++)
			*c=(BBH_D)j;
	}
	ret=gft_write_full(func_name,time,shape,cnames,rank,coords,data);
	if(coords)
		free(coords);
	for(i=0;i<rank;i++)
		if(cnames[i])
			free(cnames[i]);
	if(cnames)
		free(cnames);
	return(ret);
}

BBH_I gft_write_bbox(const BBH_C *func_name, /* name of this grid function */
										BBH_D time, /* time of this slice */
										BBH_I *shape, /* shape of data set */
										BBH_I rank, /* rank of data set */
										BBH_D *coords, /* values of coordinate bboxes */
										BBH_D *data ) /* actual data */
{
	BBH_C **cnames;
	BBH_D *ncoords,*nc,dc;
	BBH_I i,j,sz,ret;
	
	if(rank<=0){
		return(gft_write_full(func_name,time,shape,cnames,rank,ncoords,data));
	}
	if((cnames=(BBH_C **)malloc(sizeof(BBH_C *)*(rank)))==NULL){
		fprintf(stderr,"gft_write: can't get memory for coordinate names.\n");
		errno=9;
		return(0);
	}
	for(i=0;i<rank;i++){
		if((cnames[i]=(BBH_C *)malloc(sizeof(BBH_C)*2))==NULL){
			fprintf(stderr,"gft_write: can't get memory for coordinate name.\n");
			errno=10;
			return(0);
		}
	}
	for(sz=0,i=0;i<rank;i++){
		cnames[i][0]='x'+i;
		cnames[i][1]=0;
		sz+=shape[i];
	}
	if((ncoords=(BBH_D *)malloc(sizeof(BBH_D)*sz))==NULL){
		fprintf(stderr,"gft_write: can't get memory for coordinates.\n");
		errno=11;
		return(0);
	}
	for(nc=ncoords,i=0;i<rank;i++){
		dc=(coords[2*i+1]-coords[2*i])/(shape[i]-1);
		for(j=0;j<shape[i];j++,nc++)
			*nc=j*dc+coords[2*i];
	}
	ret=gft_write_full(func_name,time,shape,cnames,rank,ncoords,data);
	if(ncoords)
		free(ncoords);
	for(i=0;i<rank;i++)
		if(cnames[i])
			free(cnames[i]);
	if(cnames)
		free(cnames);
	return(ret);
}

/* rank <=0 and func_name==NULL means close all open files */
/* rank <=0 means close file */
BBH_I gft_write_full(const BBH_C *func_name, /* name of this grid function */
										BBH_D time, /* time of this slice */
										BBH_I *shape, /* shape of data set */
										BBH_C **cnames, /* names of coordinates */
										BBH_I rank, /* rank of data set */
										BBH_D *coords, /* values of coordinates */
										BBH_D *data ) /* actual data */
{
  BBH_I i,j,k,ind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id,dim_id;
  BBH_C name[60],cnm[50],stp[12],*nm;
  BBH_I mode,nslabs,gattrs;
	BBH_D *crds;
	BBH_I ltrace=0;
	
	if(ltrace){
		fprintf(stderr,"gft_write_full: call=%d top=%d\n",call,top);
	}
		
	if(rank<=0){ /* close all files and return */
		if(ltrace)
			fprintf(stderr,"gft_write_full: closing files\n");
		if(func_name == NULL){
			gft_close_all();
		}else{
			nm=gft_make_file_name(func_name);
			gft_close(nm);
		}
		return(1);
	}
	nm=gft_make_file_name(func_name);
	if(ltrace)
		fprintf(stderr,"gft_write_full: nm=<%s>\n",nm);

	ind=gft_create(nm);
	if(ind==-1){
		fprintf(stderr,"gft_write_full: unable to open file <%s>\n",nm);
		errno=1;
		return(0);
	}
	if(ltrace)
		fprintf(stderr,"gft_write_full: ind=%d\n",ind);
		
	fid=dfiles[ind].fid;
	if(SDfileinfo(fid,(int32 *)&nslabs,(int32 *)&gattrs)==-1){
		fprintf(stderr,"gft_write_full: Can't get file info.\n");
		errno=0;
		return(0);
	}
	dfiles[ind].sdsnum=(nslabs-rank > 0)?nslabs-rank:0;
	
	if(ltrace)
		fprintf(stderr,"gft_write_full: rank=%d fid=%d\n",rank,fid);
  for(i=0;i<rank;start[i++]=0);
	sprintf(name,"%s[%d]",func_name,dfiles[ind].sdsnum);
	if(ltrace)
		fprintf(stderr,"gft_write_full: name=<%s>\n",name);
	if((sds_id=SDcreate(fid,name,d_type_,rank,(int32 *)shape))==-1){
		fprintf(stderr,"gft_write: can't create <%s>\n",name);
		errno=2;
		return(0);
	}
	if(ltrace)
		fprintf(stderr,"gft_write_full: created sds\n");
	if(SDwritedata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)data)==-1){
		fprintf(stderr,"gft_write: can't write data to <%s>\n",name);
		errno=3;
		return(0);
	}
	if(ltrace)
		fprintf(stderr,"gft_write_full: wrote data\n");
	if(SDsetattr(sds_id,"t",DFNT_FLOAT64,1,(BBH_V *)&time)==-1){
		fprintf(stderr,"gft_write: can't write time to <%s>\n",name);
		errno=4;
		return(0);
	}
	if(ltrace)
		fprintf(stderr,"gft_write_full: set time\n");
	for(crds=coords,k=0;k<rank;k++){
		if((dim_id=SDgetdimid(sds_id,k))==-1){
			fprintf(stderr,"gft_write: <%s>, can't get id for dim %d\n",name,k);
			errno=5;
			return(0);
		}
		if(ltrace)
			fprintf(stderr,"gft_write_full: dim_id=%d k=%d\n",dim_id,k);
		sprintf(cnm,"%s[%d]",(BBH_C *)cnames[k],dim_id);
		if(ltrace)
			fprintf(stderr,"gft_write_full: cnm=<%s>\n",cnm);
		if(SDsetdimname(dim_id,cnm)==-1){
			fprintf(stderr,"gft_write: <%s>, can't set name for dim %d\n",name,k);
			errno=6;
			return(0);
		}
		if(ltrace){
			fprintf(stderr,"gft_write_full: cnames[%d]=<%s>\n",k,cnames[k]);
			fprintf(stderr,"  shape[%d]=%d\n",k,shape[k]);
		}
		if(SDsetdimscale(dim_id,shape[k],DFNT_FLOAT64,(BBH_V *)crds)==-1){
			fprintf(stderr,"gft_write: <%s>, can't set scale for dim %d\n",name,k);
			errno=7;
			return(0);
		}
		crds+=shape[k];
	}
	if(ltrace)
		fprintf(stderr,"gft_write_full: exiting\n");
	SDendaccess(sds_id);
	if(nm) free(nm);
	return(1);
}

/* for these three read routines, storage must be pre-allocated */
BBH_I gft_read_shape(const BBH_C *file_name, /* IN: */
											const BBH_I level, /* IN: */
											BBH_I *shape) /* OUT: */
{
	BBH_C *nm;
	BBH_I ind,fid,sds_id,sind;
  BBH_I type,rank,attrs,nslabs,gattrs;
  BBH_C name[60];
	
	nm=gft_make_file_name(file_name);
	ind=gft_open_rd(nm);
	if(ind==-1){
		fprintf(stderr,"gft_read_shape: unable to open file <%s>\n",nm);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
	if(dfiles[ind].rank==0){ /* initialize */
		if(SDfileinfo(fid,(int32 *)&nslabs,(int32 *)&gattrs)==-1){
			fprintf(stderr,"gft_read_shape: Can't get file info.\n");
			errno=2;
			return(0);
		}
		dfiles[ind].nslabs=nslabs;
	  if((sds_id=SDselect(fid,0))==-1){
	    fprintf(stderr,"gft_read_shape: Can't select 1st sds.\n");
	    errno=3;
	    return(0);
	  }
	  if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)shape,(int32 *)&type,
								(int32 *)&attrs)==-1){
	    fprintf(stderr,"gft_read_shape: Can't get info about 1st sds\n");
	    errno=4;
	    return(0);
	  }
		dfiles[ind].rank=rank;
		if(type!=DFNT_FLOAT64){
			fprintf(stderr,"gft_read_shape: data set type is wrong.\n");
			errno=5;
			return(0);
		}
	}
	rank=dfiles[ind].rank;
	if(dfiles[ind].vers==2){
		sind=(level-1)*(rank+1);
	}else if(dfiles[ind].vers==1){
		sind=(level==1)?0:level+rank-1;
	}
	if((sds_id=SDselect(fid,sind))==-1){
		fprintf(stderr,"gft_read_shape: Can't select sds %d.\n",(level-1)*(rank+1));
		errno=6;
		return(0);
	}
	if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)shape,(int32 *)&type,
							(int32 *)&attrs)==-1){
		fprintf(stderr,"gft_read_shape: Can't get info about sds %d\n",(level-1)*(rank+1));
		errno=7;
		return(0);
	}
	return(1);
}

BBH_I gft_read_full(const BBH_C *file_name, /* IN: name of this file */
										BBH_I level, /* IN: time level */
										BBH_I *shape, /* OUT: shape of data set */
										BBH_C **cnames, /* OUT: names of coordinates */
										BBH_I rank, /* IN: rank of data set */
										BBH_D *time, /* OUT: time value */
										BBH_D *coords, /* OUT: values of coordinates */
										BBH_D *data ) /* OUT: actual data */
{
  BBH_I i,j,k,ind,type,sind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id,dim_id,nlev;
  BBH_C name[60],*nm,*strip;
  BBH_I nslabs,gattrs,attrs,sz,dattrs;
	BBH_D *crds,t;
	BBH_I ltrace=0;
	
	if(ltrace){
		fprintf(stderr,"gft_read_full: call=%d top=%d\n",call,top);
	}
		
	nm=gft_make_file_name(file_name);
	ind=gft_open_rd(nm);
	if(ind==-1){
		fprintf(stderr,"gft_read_full: unable to open file <%s>\n",nm);
		errno=1;
		return(0);
	}
	
	if(ltrace)
		fprintf(stderr,"gft_read_full: ind=%d\n",ind);
		
	fid=dfiles[ind].fid;
	if(ltrace)
		fprintf(stderr,"gft_read_full: rank=%d fid=%d\n",rank,fid);

	if(dfiles[ind].rank==0){ /* initialize */
		if(SDfileinfo(fid,(int32 *)&nslabs,(int32 *)&gattrs)==-1){
			fprintf(stderr,"gft_read_full: Can't get file info.\n");
			errno=2;
			return(0);
		}
		dfiles[ind].nslabs=nslabs;
	  if((sds_id=SDselect(fid,0))==-1){
	    fprintf(stderr,"gft_read_full: Can't select 1st sds.\n");
	    errno=3;
	    return(0);
	  }
	  if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)shape,(int32 *)&type,
								(int32 *)&attrs)==-1){
	    fprintf(stderr,"gft_read_full: Can't get info about 1st sds\n");
	    errno=4;
	    return(0);
	  }
		dfiles[ind].rank=rank;
		if(type!=DFNT_FLOAT64){
			fprintf(stderr,"gft_read_full: data set type is wrong.\n");
			errno=5;
			return(0);
		}
	}
	rank=dfiles[ind].rank;
	nslabs=dfiles[ind].nslabs;
	if(dfiles[ind].vers==2){
		sind=(level-1)*(rank+1);
		nlev=nslabs/(rank+1);
	}else if(dfiles[ind].vers==1){
		sind=(level==1)?0:level+rank-1;
		nlev=nslabs-rank;
	}
	if(level>nlev || level<1){
		fprintf(stderr,"gft_read_full: time level %d out of range.\n",level);
		errno=6;
		return(0);
	}
  for(i=0;i<rank;i++){
		start[i]=0;
	}
	if((sds_id=SDselect(fid,sind))==-1){
		fprintf(stderr,"gft_read_full: Can't select sds number %d.\n",sind);
		errno=12;
		return(0);
	}
	if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)shape,(int32 *)&type,
							(int32 *)&attrs)==-1){
		fprintf(stderr,"gft_read_full: Can't get info about sds %d\n",sind);
		errno=13;
		return(0);
	}
	for(crds=coords,i=0;i<rank;i++){
		dim_id=SDgetdimid(sds_id,i);
		if(dim_id!=-1){
			if(SDdiminfo(dim_id,name,(int32 *)&sz,(int32 *)&type,(int32 *)&dattrs)==-1){
				fprintf(stderr,"gft_extract: Can't get dimension info.\n");
				errno=18;
				return(0);
			}
			if(type==0){
				fprintf(stderr,"gft_read_full: Bad dimension %d.\n",i);
				errno=19;
				return(0);
			}
			strip=strrchr(name,'[');
			if(strip){
				strncpy(cnames[i],name,(BBH_I)(strip-name));
				cnames[i][(BBH_I)(strip-name)]=0;
			}else{
				strcpy(cnames[i],name);
			}
			if(sz!=shape[i]){
				fprintf(stderr,"gft_read_full: Dim %d has size %d while shape=%d.\n",
								i,sz,shape[i]);
				errno=21;
				return(0);
			}
			if(SDgetdimscale(dim_id,crds)==-1){
				fprintf(stderr,"gft_read_full: Can't read dim scale %d.\n",i);
				errno=23;
				return(0);
			}
		}else{
			fprintf(stderr,"gft_read_full: Can't get dim id %d.\n",i);
			errno=24;
			return(0);
		}
		crds+=shape[i];
	}
	if(SDreadattr(sds_id,0,&t)==-1){
		fprintf(stderr,"gft_read_full: Can't get time value.\n");
		errno=25;
		return(0);
	}
	*time=t;
	if(SDreaddata(sds_id,(int32 *)start,NULL,(int32 *)shape,
								(BBH_V *)(data))==-1){
		fprintf(stderr,"gft_read_full: Can't read data.\n");
		errno=27;
		return(0);
	}
	SDendaccess(sds_id);
	return 1;
}

BBH_I gft_read_brief(const BBH_C *file_name, /* IN: name of this file */
										BBH_I level, /* IN: time level */
										BBH_D *data ) /* OUT: actual data */
{
  BBH_I i,j,k,ind,type,sind;
  BBH_I start[MAXRANK],size[MAXRANK];
  BBH_I fid,sds_id,dim_id,rank,nlev;
  BBH_C name[60],*nm;
  BBH_I nslabs,gattrs,attrs,sz,dattrs;
	BBH_D *crds,t;
	BBH_I ltrace=0;
	
	if(ltrace){
		fprintf(stderr,"gft_read_brief: call=%d top=%d\n",call,top);
	}
		
	nm=gft_make_file_name(file_name);
	ind=gft_open_rd(nm);
	if(ind==-1){
		fprintf(stderr,"gft_read_brief: unable to open file <%s>\n",nm);
		errno=1;
		return(0);
	}
	
	if(ltrace)
		fprintf(stderr,"gft_read_brief: ind=%d\n",ind);
		
	fid=dfiles[ind].fid;
	if(ltrace)
		fprintf(stderr,"gft_read_brief: fid=%d\n",fid);

	if(dfiles[ind].rank==0){ /* initialize */
		if(SDfileinfo(fid,(int32 *)&nslabs,(int32 *)&gattrs)==-1){
			fprintf(stderr,"gft_read_brief: Can't get file info.\n");
			errno=2;
			return(0);
		}
		dfiles[ind].nslabs=nslabs;
	  if((sds_id=SDselect(fid,0))==-1){
	    fprintf(stderr,"gft_read_brief: Can't select 1st sds.\n");
	    errno=3;
	    return(0);
	  }
	  if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)size,(int32 *)&type,
								(int32 *)&attrs)==-1){
	    fprintf(stderr,"gft_read_brief: Can't get info about 1st sds\n");
	    errno=4;
	    return(0);
	  }
		dfiles[ind].rank=rank;
		if(type!=DFNT_FLOAT64){
			fprintf(stderr,"gft_read_brief: data set type is wrong.\n");
			errno=5;
			return(0);
		}
	}
	rank=dfiles[ind].rank;
	nslabs=dfiles[ind].nslabs;
	if(dfiles[ind].vers==2){
		sind=(level-1)*(rank+1);
		nlev=nslabs/(rank+1);
	}else if(dfiles[ind].vers==1){
		sind=(level==1)?0:level+rank-1;
		nlev=nslabs-rank;
	}
	if(level>nlev || level<1){
		fprintf(stderr,"gft_read_brief: time level %d out of range.\n",level);
		errno=6;
		return(0);
	}
  for(i=0;i<rank;i++){
		start[i]=0;
	}
	if((sds_id=SDselect(fid,sind))==-1){
		fprintf(stderr,"gft_read_brief: Can't select sds number %d.\n",sind);
		errno=12;
		return(0);
	}
	if(SDgetinfo(sds_id,name,(int32 *)&rank,(int32 *)size,(int32 *)&type,
							(int32 *)&attrs)==-1){
		fprintf(stderr,"gft_read_brief: Can't get info about sds %d\n",sind);
		errno=13;
		return(0);
	}
	if(SDreaddata(sds_id,(int32 *)start,NULL,(int32 *)size,
								(BBH_V *)(data))==-1){
		fprintf(stderr,"gft_read_brief: Can't read data.\n");
		errno=27;
		return(0);
	}
	SDendaccess(sds_id);
	return 1;
}

/* initial data support */

BBH_I gft_read_id_str_p(const BBH_C *file_name, const BBH_C *param_name,
												BBH_C **param, BBH_I nparam)
{
  BBH_I aind,ind,fid,i;
  BBH_C name[256],*nm;
	
	nm=strdup(file_name);
	if((ind=gft_open_rd(nm))==-1){
		fprintf(stderr,"gft_read_id_str_p: can't open file <%s>\n",nm);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
	for(i=0;i<nparam;i++){
		sprintf(name,"%s[%d]",param_name,i);
		if((aind=SDfindattr(fid,name))==-1){
			fprintf(stderr,"gft_read_id_str_p: can't find attribute <%s>\n",name);
			errno=2;
			return(0);
		}
		if(SDreadattr(fid,aind,param[i])==-1){
			fprintf(stderr,"gft_read_id_str_p: can't read attribute <%s>\n",name);
			errno=2;
			return(0);
		}
	}
}

BBH_I gft_read_id_int_p(const BBH_C *file_name, const BBH_C *param_name,
												BBH_I *param, BBH_I nparam)
{
  BBH_I aind,ind,fid;
  BBH_C *nm;
	
	nm=strdup(file_name);
	if((ind=gft_open_rd(nm))==-1){
		fprintf(stderr,"gft_read_id_int_p: can't open file <%s>\n",nm);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
	if((aind=SDfindattr(fid,(BBH_C *)param_name))==-1){
		fprintf(stderr,"gft_read_id_int_p: can't find attribute <%s>\n",param_name);
		errno=2;
		return(0);
	}
	if(SDreadattr(fid,aind,param)==-1){
		fprintf(stderr,"gft_read_id_int_p: can't read attribute <%s>\n",param_name);
		errno=3;
		return(0);
	}
}

BBH_I gft_read_id_float_p(const BBH_C *file_name, const BBH_C *param_name,
												BBH_D *param, BBH_I nparam)
{
  BBH_I aind,ind,fid;
  BBH_C *nm;
	
	nm=strdup(file_name);
	if((ind=gft_open_rd(nm))==-1){
		fprintf(stderr,"gft_read_id_float_p: can't open file <%s>\n",nm);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
	if((aind=SDfindattr(fid,(BBH_C *)param_name))==-1){
		fprintf(stderr,"gft_read_id_float_p: can't find attribute <%s>\n",param_name);
		errno=2;
		return(0);
	}
	if(SDreadattr(fid,aind,param)==-1){
		fprintf(stderr,"gft_read_id_float_p: can't read attribute <%s>\n",param_name);
		errno=3;
		return(0);
	}
}

BBH_I gft_read_2idata(const BBH_C *file_name, /* name of idata file */
										const BBH_C *func_name, /* name of this grid function */
										BBH_I *shape, /* shape of data set */
										BBH_I rank, /* rank of data set */
										BBH_D *datanm1,
										BBH_D *datan ) /* actual data */
{
  BBH_I i,ind,sdind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id;
  BBH_C name[60];
	BBH_I ltrace=0;
	
	if((ind=gft_open_rd(file_name))==-1){
		fprintf(stderr,"gft_read_2idata: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
  for(i=0;i<rank;start[i++]=0);
	sprintf(name,"%s[0]",func_name);
	if((sdind=SDnametoindex(fid,name))==-1){
		fprintf(stderr,"gft_read_2idata: unable to find <%s> in <%s>\n",name,file_name);
		errno=2;
		return(0);
	}
	if((sds_id=SDselect(fid,sdind))==-1){
		fprintf(stderr,"gft_read_2idata: can't select sds\n");
		errno=3;
		return(0);
	}
	if(SDreaddata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)datanm1)==-1){
		fprintf(stderr,"gft_read_2idata: can't read data from <%s>\n",name);
		errno=4;
		return(0);
	}
	SDendaccess(sds_id);
	sprintf(name,"%s[1]",func_name);
	if((sdind=SDnametoindex(fid,name))==-1){
		fprintf(stderr,"gft_read_2idata: unable to find <%s> in <%s>\n",name,file_name);
		errno=2;
		return(0);
	}
	if((sds_id=SDselect(fid,sdind))==-1){
		fprintf(stderr,"gft_read_2idata: can't select sds\n");
		errno=3;
		return(0);
	}
	if(SDreaddata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)datan)==-1){
		fprintf(stderr,"gft_read_2idata: can't read data from <%s>\n",name);
		errno=4;
		return(0);
	}
	SDendaccess(sds_id);
	return(1);
}

BBH_I gft_read_1idata(const BBH_C *file_name, /* name of idata file */
										const BBH_C *func_name, /* name of this grid function */
										BBH_I *shape, /* shape of data set */
										BBH_I rank, /* rank of data set */
										BBH_D *datan ) /* actual data */
{
  BBH_I i,ind,sdind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id;
  BBH_C name[60];
	BBH_I ltrace=0;
	
	if((ind=gft_open_rd(file_name))==-1){
		fprintf(stderr,"gft_read_1idata: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
  for(i=0;i<rank;start[i++]=0);
	sprintf(name,"%s[0]",func_name);
	if((sdind=SDnametoindex(fid,name))==-1){
		fprintf(stderr,"gft_read_1idata: unable to find <%s> in <%s>\n",name,file_name);
		errno=2;
		return(0);
	}
	if((sds_id=SDselect(fid,sdind))==-1){
		fprintf(stderr,"gft_read_1idata: can't select sds\n");
		errno=3;
		return(0);
	}
	if(SDreaddata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)datan)==-1){
		fprintf(stderr,"gft_read_1idata: can't read data from <%s>\n",name);
		errno=4;
		return(0);
	}
	SDendaccess(sds_id);
	return(1);
}

BBH_I gft_read_idata(const BBH_C *file_name, /* name of idata file */
										const BBH_C *func_name, /* name of this grid function */
										BBH_I *shape, /* shape of data set */
										BBH_I rank, /* rank of data set */
										BBH_D *data) /* actual data */
{
  BBH_I i,ind,sdind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id;
	BBH_I ltrace=0;
	
	if((ind=gft_open_rd(file_name))==-1){
		fprintf(stderr,"gft_read_idata: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
  for(i=0;i<rank;start[i++]=0);
	if((sdind=SDnametoindex(fid,(BBH_C *)func_name))==-1){
		fprintf(stderr,"gft_read_idata: unable to find <%s> in <%s>\n",func_name,file_name);
		errno=2;
		return(0);
	}
	if((sds_id=SDselect(fid,sdind))==-1){
		fprintf(stderr,"gft_read_idata: can't select sds\n");
		errno=3;
		return(0);
	}
	if(SDreaddata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)data)==-1){
		fprintf(stderr,"gft_read_1idata: can't read data from <%s>\n",func_name);
		errno=4;
		return(0);
	}
	SDendaccess(sds_id);
	return(1);
}

BBH_I gft_write_id_str_p(const BBH_C *file_name, const BBH_C *param_name,
												BBH_C **param, BBH_I nparam)
{
  BBH_I ind,fid,i;
  BBH_C name[256];
	
	if((ind=gft_create(file_name))==-1){
		fprintf(stderr,"gft_write_id_str_p: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
	for(i=0;i<nparam;i++){
		sprintf(name,"%s[%d]",param_name,i);
		if(SDsetattr(fid,name,DFNT_CHAR8,strlen(param[i])+1,param[i])==-1){
			fprintf(stderr,"gft_write_id_str_p: can't set attribute <%s>\n",name);
			errno=2;
			return(0);
		}
	}
}

BBH_I gft_write_id_int_p(const BBH_C *file_name, const BBH_C *param_name,
												BBH_I *param, BBH_I nparam)
{
  BBH_I ind,fid;
	
	if((ind=gft_create(file_name))==-1){
		fprintf(stderr,"gft_write_id_int_p: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
	if(SDsetattr(fid,(BBH_C *)param_name,DFNT_INT32,nparam,param)==-1){
		fprintf(stderr,"gft_write_id_int_p: can't set attribute <%s>\n",param_name);
		errno=2;
		return(0);
	}
}

BBH_I gft_write_id_float_p(const BBH_C *file_name, const BBH_C *param_name,
												BBH_D *param, BBH_I nparam)
{
  BBH_I ind,fid;
	
	if((ind=gft_create(file_name))==-1){
		fprintf(stderr,"gft_write_id_float_p: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
	if(SDsetattr(fid,(BBH_C *)param_name,DFNT_FLOAT64,nparam,param)==-1){
		fprintf(stderr,"gft_write_id_float_p: can't set attribute <%s>\n",param_name);
		errno=2;
		return(0);
	}
}

BBH_I gft_write_2idata(const BBH_C *file_name, /* name of idata file */
									const	BBH_C *func_name, /* name of this grid function */
										BBH_I *shape, /* shape of data set */
										BBH_I rank, /* rank of data set */
										BBH_D *datanm1,
										BBH_D *datan ) /* actual data */
{
  BBH_I i,ind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id;
  BBH_C name[60];
	BBH_I ltrace=0;
	
	if((ind=gft_create(file_name))==-1){
		fprintf(stderr,"gft_write_2idata: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
  for(i=0;i<rank;start[i++]=0);
	sprintf(name,"%s[0]",func_name);
	if((sds_id=SDcreate(fid,name,d_type_,rank,(int32 *)shape))==-1){
		fprintf(stderr,"gft_write_2idata: can't create %s\n",name);
		errno=2;
		return(0);
	}
	if(SDwritedata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)datanm1)==-1){
		fprintf(stderr,"gft_write_2idata: can't write data to %s\n",name);
		errno=3;
		return(0);
	}
	SDendaccess(sds_id);
	sprintf(name,"%s[1]",func_name);
	if((sds_id=SDcreate(fid,name,d_type_,rank,(int32 *)shape))==-1){
		fprintf(stderr,"gft_write_2idata: can't create %s\n",name);
		errno=2;
		return(0);
	}
	if(SDwritedata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)datan)==-1){
		fprintf(stderr,"gft_write_2idata: can't write data to %s\n",name);
		errno=3;
		return(0);
	}
	SDendaccess(sds_id);
	return(1);
}

BBH_I gft_write_1idata(const BBH_C *file_name, /* name of idata file */
									const	BBH_C *func_name, /* name of this grid function */
										BBH_I *shape, /* shape of data set */
										BBH_I rank, /* rank of data set */
										BBH_D *datan ) /* actual data */
{
  BBH_I i,ind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id;
  BBH_C name[60];
	BBH_I ltrace=0;
	
	if((ind=gft_create(file_name))==-1){
		fprintf(stderr,"gft_write_1idata: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
  for(i=0;i<rank;start[i++]=0);
	sprintf(name,"%s[0]",func_name);
	if((sds_id=SDcreate(fid,name,d_type_,rank,(int32 *)shape))==-1){
		fprintf(stderr,"gft_write_1idata: can't create %s\n",name);
		errno=2;
		return(0);
	}
	if(SDwritedata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)datan)==-1){
		fprintf(stderr,"gft_write_1idata: can't write data to %s\n",name);
		errno=3;
		return(0);
	}
	SDendaccess(sds_id);
	return(1);
}

BBH_I gft_write_idata(const BBH_C *file_name, /* name of idata file */
										const BBH_C *func_name, /* name of this grid function */
										BBH_I *shape, /* shape of data set */
										BBH_I rank, /* rank of data set */
										BBH_D *data) /* actual data */
{
  BBH_I i,ind;
  BBH_I start[MAXRANK];
  BBH_I fid,sds_id;
	BBH_I ltrace=0;
	
	if((ind=gft_create(file_name))==-1){
		fprintf(stderr,"gft_write_idata: can't open file <%s>\n",file_name);
		errno=1;
		return(0);
	}
	fid=dfiles[ind].fid;
  for(i=0;i<rank;start[i++]=0);
	if((sds_id=SDcreate(fid,(BBH_C *)func_name,d_type_,rank,(int32 *)shape))==-1){
		fprintf(stderr,"gft_write_idata: can't create %s\n",func_name);
		errno=2;
		return(0);
	}
	if(SDwritedata(sds_id,(int32 *)start,NULL,(int32 *)shape,(BBH_V *)data)==-1){
		fprintf(stderr,"gft_write_1idata: can't write data to %s\n",func_name);
		errno=3;
		return(0);
	}
	SDendaccess(sds_id);
	return(1);
}
